#version 330 core
const int MAX_LIGHTS = 10;

in vec2 fUVs;
in vec3 Position_worldspace;
in vec3 Normal_worldspace;
//in vec3 Normal_cameraspace;
in vec3 EyeDirection_cameraspace;
in vec3 LightDirection_cameraspace[MAX_LIGHTS];
in vec4 ShadowCoord;
in vec3 tangentout;

out vec3 color;

uniform sampler2D myTextureSampler;
uniform sampler2D normalMap;
uniform sampler2D shadowMap;
uniform vec3 LightInvDirection;

uniform mat4 modelMatrix;
uniform int NUM_LIGHTS;
uniform int lightType[MAX_LIGHTS];
uniform vec3 lightPosition[MAX_LIGHTS];
uniform vec3 lightDirection[MAX_LIGHTS];
uniform	float cutoff[MAX_LIGHTS];
uniform	float exponent[MAX_LIGHTS];
uniform vec3 lightPower[MAX_LIGHTS];
uniform mat4 dV;
uniform mat4 lightProj;

vec3 CalcBumpedNormal()                                                                     
{                                                                                           
    vec3 Normal = normalize(Normal_worldspace);                                                       
    vec3 Tangent = normalize(tangentout);                                                     
    Tangent = normalize(Tangent - dot(Tangent, Normal) * Normal);                           
    vec3 Bitangent = cross(Tangent, Normal);                                                
    vec3 BumpMapNormal = texture(normalMap, fUVs).xyz;                                
    BumpMapNormal = 2.0 * BumpMapNormal - vec3(1.0, 1.0, 1.0);                              
    vec3 NewNormal;                                                                         
    mat3 TBN = mat3(Tangent, Bitangent, Normal);                                            
    NewNormal = TBN * BumpMapNormal;                                                        
    NewNormal = normalize(NewNormal);                                                       
    return NewNormal;                                                                       
}

float ChebychevInequality (vec2 moments, float t) {
	// No shadow if depth of fragment is in front
	if ( t <= moments.x )
		return 1.0;

	// Calculate variance, which is actually the amount of
	// error due to precision loss from fp32 to RG/BA
	// (moment1 / moment2)
	float variance = moments.y - (moments.x * moments.x);
	variance = max(variance, 0.002);
	
	// Calculate the upper bound
	float d = t - moments.x;
	
	return (variance / (variance + d * d));
}

float VsmFixLightBleed (float pMax, float amount) {
	return clamp((pMax - amount) / (1.0 - amount), 0.0, 1.0);
}
float linstep(float low, float high, float v){
	return clamp((v-low)/(high-low), 0.0, 1.0);
}

float VSM(sampler2D depths, vec2 uv, float compare){
	vec2 moments = texture(depths, uv).xy;
	float p = smoothstep(compare-0.02, compare, moments.x);
	float variance = max(moments.y - moments.x*moments.x, -0.001);
	float d = compare - moments.x;
	float p_max = linstep(0.2, 1.0, variance / (variance + d*d));
	return clamp(max(p, p_max), 0.0, 1.0);
}

void main(){

	
	// Material properties
	vec3 MaterialDiffuseColor = texture( myTextureSampler, fUVs ).rgb;
	vec3 MaterialAmbientColor = vec3(0.1,0.1,0.1) * MaterialDiffuseColor;
	vec3 MaterialSpecularColor = vec3(0.7,0.7,0.7);

	vec3 lightDiffuse = vec3(0,0,0);
	vec3 lightSpecular = vec3(0,0,0);
	float lightSpot = 0;


	vec3 n = CalcBumpedNormal();
	
	for(int i = 0; i < NUM_LIGHTS; ++i){

		if(lightType[i] == 0) // Ambient
		{
			lightDiffuse +=lightPower[i];
		}
		else if(lightType[i] == 1) // Spot
		{

			vec3 lightVec =  Position_worldspace-lightPosition[i];
			float lightDist = length(lightPosition[i]-Position_worldspace);
			// Normal of the computed fragment, in camera space
			//vec3 n = normalize( Normal_cameraspace );
			// Direction of the light (from the fragment to the light)
			vec3 l = normalize( LightDirection_cameraspace[i] );
			// Eye vector (towards the camera)
			vec3 E = normalize(EyeDirection_cameraspace);
 			// Direction in which the triangle reflects the light
			vec3 R = reflect(-l,n);
			float cosAlpha = clamp( dot( E,R ), 0,1 );
			float angleDot = dot(normalize(lightDirection[i]),normalize(lightVec));
			float lightAngleCos = cos(cutoff[i]*2*3.1415/360);
			if (angleDot>lightAngleCos) {
				//lightDiffuse += max(1-lightDist / 35, 0)*lightPower[i];
				float fDif = 1.0-lightAngleCos;
				float fFactor = clamp((angleDot-lightAngleCos)/fDif, 0.0, 1.0); 
				
				//lightSpot += max(1-lightDist / 35, 0)*lightPower[i];
				lightSpot = fFactor/(lightDist*0.3); 
				lightSpecular += pow(cosAlpha,5) / (lightDist*lightDist)*lightPower[i]*20;
				//lightSpot += 1;
			}
		}
		else if(lightType[i] == 2) // Point
		{
			float distance = length( lightPosition[i] - Position_worldspace );

			// Normal of the computed fragment, in camera space
			//vec3 n = normalize( Normal_cameraspace );
			// Direction of the light (from the fragment to the light)
			vec3 l = normalize( LightDirection_cameraspace[i] );
			// Cosine of the angle between the normal and the light direction, 
			// clamped above 0
			//  - light is at the vertical of the triangle -> 1
			//  - light is perpendicular to the triangle -> 0
			//  - light is behind the triangle -> 0
			float cosTheta = clamp( dot( n,l ), 0,1 );
	
			// Eye vector (towards the camera)
			vec3 E = normalize(EyeDirection_cameraspace);
 			// Direction in which the triangle reflects the light
			vec3 R = reflect(-l,n);
			// Cosine of the angle between the Eye vector and the Reflect vector,
			// clamped to 0
			//  - Looking into the reflection -> 1
			//  - Looking elsewhere -> < 1
			float cosAlpha = clamp( dot( E,R ), 0,1 );
			lightDiffuse += (cosTheta / (distance*distance))*lightPower[i];
			lightSpecular += pow(cosAlpha,5) / (distance*distance)*lightPower[i];

		}
		else if(lightType[i] == 3) // Directional
		{
			
		}


	}
	vec3 depth = ShadowCoord.xyz / ShadowCoord.w;
	float shadow = 1.0;

	// shadow calculation
	vec3 Position_lightSpace = (dV * vec4(Position_worldspace,1)).xyz;
	
    vec4 lightDevice = lightProj * vec4(Position_lightSpace, 1.0);
    vec2 lightDeviceNormal = lightDevice.xy/lightDevice.w;
    vec2 lightUV = lightDeviceNormal*0.5+0.5;
	
    float lightDepth2 = clamp(length(Position_lightSpace)/22.0, 0.0, 1.0);
    float illuminated = VSM(shadowMap, depth.xy, lightDepth2);
	
	vec3 tmpColor = 
		MaterialAmbientColor +
		MaterialDiffuseColor  * lightDiffuse +
		illuminated * MaterialDiffuseColor * lightSpot +
		MaterialSpecularColor * lightSpecular;
	color = tmpColor;
	color = clamp(tmpColor /** illuminated*/, 0.0, 1.0);
	color = pow(color,vec3(2.2)); 	//clamp(vec3(1.0,1.0,1.0)*shadow,0.0,1.0);
	//color = textureProj(shadowMap, ShadowCoord) * tmpColor;

		/*		
		color = 
		MaterialAmbientColor +
		MaterialDiffuseColor * LightColor * LightPower * cosTheta / (distance*distance) +
		MaterialSpecularColor * LightColor * LightPower * pow(cosAlpha,5) / (distance*distance);
		*/
		
}